﻿using System;

namespace TestClient {
  class Program {
    const int NumberOfRequestedMessages = 500;
    static int NumberOfReceivedMessages = 0;
    static System.Threading.ManualResetEvent AllMessagesReceived = new System.Threading.ManualResetEvent(false);

    static void Main(string[] args)
    {
      // Initialize the matrix
      zx.matrix.CMatrixSubsystem.Start("TestClient", "1.1");

      // Subscribe for test messages sent by the server
      // This subscription is "not blocking": it guarantees that MsgHandler_TestMsg will never
      // be called in race condition
      zx.matrix.CMatrixSubsystem.Subscribe(
        new zx.matrix.CStdMessage("TestMsg", "Test"),
        new zx.matrix.ProcessHandle(MsgHandler_TestMsg));

      // Connection events examples
      // This is called when the component has connected to the node
      zx.matrix.CMatrixSubsystem.Subscribe(
        new zx.matrix.CEventConnected(),
        (msg, addr) =>
          Console.WriteLine(
            "CONNECTED HAS TO " + zx.matrix.CMatrixSubsystem.GetMasterAddress().GetFullName()),
        zx.matrix.CMatrixSubsystem.ESubscriptionType.Event);
      // This is called when the component has disconnected from the node
      zx.matrix.CMatrixSubsystem.Subscribe(
        new zx.matrix.CEventDisconnected(),
        (msg, addr) => Console.WriteLine("DISCONNECTED"),
        zx.matrix.CMatrixSubsystem.ESubscriptionType.Event);

      // Wait for server initialization (with literaly unlimited timeout)
      Console.WriteLine("Waiting for TestServer initialization...");
      zx.matrix.CMatrixSubsystem.WaitForComponentInit("TestServer", uint.MaxValue);

      // Ask server name
      Console.WriteLine("Asking server name...");
      zx.pl.CPropertyList outParams = null;
      var outError = new zx.CError();
      if (zx.matrix.CMatrixSubsystem.SynchRequest(
            aComponentName: "TestServer",
            aMethodName: "GetMyName", 
            aInParameters: new zx.pl.CBranch(),
            outParameters: ref outParams,
            outError: ref outError, 
            aPriority: zx.matrix.TPriority.PRIORITY_HIGHEST))
      {
        // Executed ok
        Console.WriteLine(
          "Server name is: " + outParams.GetProperty("MyName").DC.GetStringValue());
  	  } // if
      else
      {
        // Call ended with error
        Console.WriteLine("ERROR: " + outError.ErrorMessage);
        outError.ClearError(); // to re-use the same error object below
      }

      // Invoke server method "SendMessages"

      // Do it with invalid parameters (just for example)
      Console.WriteLine("Calling 'SendMessages' with bad input...");
      var inParams = new zx.pl.CBranch();
      inParams.AddPropertyValue("Number", 100000); // requested too many messages
      zx.matrix.CMatrixSubsystem.SynchRequest(
            aComponentName: "TestServer",
            aMethodName: "SendMessages", 
            aInParameters: inParams,
            outParameters: ref outParams,
            outError: ref outError, 
            aPriority: zx.matrix.TPriority.PRIORITY_HIGHEST);
      Console.WriteLine("ERROR: " + outError.ErrorMessage);

      // Now, call the method correctly
      Console.WriteLine("Calling 'SendMessages' with good input...");
      // request sending messages in parameter 'Number'. This parameter was added above in inParams
      inParams.GetProperty("Number").DC.SetLongValue(NumberOfRequestedMessages);
      if (zx.matrix.CMatrixSubsystem.SynchRequest(
            aComponentName: "TestServer",
            aMethodName: "SendMessages", 
            aInParameters: inParams,
            outParameters: ref outParams,
            outError: ref outError, 
            aPriority: zx.matrix.TPriority.PRIORITY_HIGHEST))
      {
        // Request sent successfully: wait till we receive all messages
        Console.WriteLine("Waiting for receiving the messages...");
        AllMessagesReceived.WaitOne();
        Console.WriteLine("All messages have been received. Mission complete, quitting...");
      }
      else
      {
        Console.WriteLine("ERROR: " + outError.ErrorMessage);
        Console.WriteLine("Disappointed, quitting...");
      }

      // Inform anybody who's interested about finishing
      var msgInfo = new zx.matrix.CStdMessage("TestMsgFinishInfo", "Test");
      msgInfo.Root.AddPropertyValue("NumberOfReceivedMessages", NumberOfReceivedMessages);
      zx.matrix.CMatrixSubsystem.Send(msgInfo);

      zx.matrix.CMatrixSubsystem.Stop();
    }


    // Message handler for test Message
    private static void MsgHandler_TestMsg(zx.matrix.CMessage aMsg, zx.matrix.CAddress aAddr)
    {
      // We use "shared resource" NumberOfReceivedMessages here, but no need to care about
      // locking, because the subscription was done as "non blocking"
      ++NumberOfReceivedMessages;

      // Decerialize the message
      var msg = zx.matrix.CStdMessage.Deserialize(aMsg);
      Console.WriteLine(
        string.Format(
          "==> Received message {0}/{1} from {2}. Info: '{3}'",
          NumberOfReceivedMessages, NumberOfRequestedMessages,
          aAddr.GetFullName(),
          msg.GetFullType(),
          msg.Root.GetProperty("Info").DC.GetStringValue()));
      if (NumberOfReceivedMessages >= NumberOfRequestedMessages)
      {
        // All messages received: flag the event
        AllMessagesReceived.Set();
      }
    }


    private static void Event_WhenConnected(zx.matrix.CMessage aMsg, zx.matrix.CAddress aAddr)
    {
      Console.WriteLine("CONNECTED HAS TO " + zx.matrix.CMatrixSubsystem.GetMasterAddress().GetFullName());
    }


    private static void Event_WhenDisconnected(zx.matrix.CMessage aMsg, zx.matrix.CAddress aAddr)
    {
      Console.WriteLine("DISCONNECTED");
    }
  }
}
